package api

import (
	"context"
	"fmt"
	"net/http"
	"path/filepath"

	casbin "github.com/casbin/casbin/v2"
	"github.com/gin-contrib/authz"
	"github.com/gin-contrib/gzip"
	"github.com/gin-contrib/requestid"
	"github.com/gin-gonic/gin"
	appchainmgr "github.com/meshplus/bitxhub-core/appchain-mgr"
	"github.com/meshplus/bitxhub-kit/crypto/asym"
	"github.com/sirupsen/logrus"

	"gitlab.33.cn/link33/sidecar/cmd/sidecar/client"
	"gitlab.33.cn/link33/sidecar/internal/repo"
	"gitlab.33.cn/link33/sidecar/internal/router"
	"gitlab.33.cn/link33/sidecar/model/pb"
	"gitlab.33.cn/link33/sidecar/pkg/json"
	"gitlab.33.cn/link33/sidecar/tool"
)

type Server struct {
	engine *gin.Engine
	config *repo.Config
	logger logrus.FieldLogger
	ctx    context.Context
	cancel context.CancelFunc
	router router.Router
}

type response struct {
	Data []byte `json:"data"`
}

func NewServer(config *repo.Config, router router.Router, logger logrus.FieldLogger) *Server {
	ctx, cancel := context.WithCancel(context.Background())

	// load the casbin model and policy from files, database is also supported.
	e, err := casbin.NewEnforcer("authz_model.conf", "authz_policy.csv")
	tool.Asset(err)
	gin.SetMode(gin.ReleaseMode)
	r := gin.Default()
	r.NoMethod(HandleNotFound)
	r.NoRoute(HandleNotFound)
	r.Use(gzip.Gzip(gzip.DefaultCompression), ErrHandler(), requestid.New(), authz.NewAuthorizer(e))
	return &Server{
		engine: r,
		config: config,
		logger: logger,
		ctx:    ctx,
		cancel: cancel,
		router: router,
	}
}

func (g *Server) Start() error {
	g.engine.Use(gin.Recovery())
	v1 := g.engine.Group("/v1")
	// g.engine.Use(gin.Recovery())
	{
		v1.POST(client.RegisterAppchainUrl, g.registerAppchain)
		v1.POST(client.UpdateAppchainUrl, g.updateAppchain)
		v1.GET(client.GetAppchainUrl, g.getAppchain)
	}

	go func() {
		go func() {
			tool.Asset(g.engine.Run(fmt.Sprintf(":%d", g.config.Port.Http)))
		}()
		<-g.ctx.Done()
	}()
	return nil
}

func (g *Server) Stop() error {
	g.cancel()
	g.logger.Infoln("gin service stop")
	return nil
}

func (g *Server) updateAppchain(c *gin.Context) {
	g.sendAppchain(c, pb.Message_APPCHAIN_UPDATE)
}

func (g *Server) registerAppchain(c *gin.Context) {
	g.sendAppchain(c, pb.Message_APPCHAIN_REGISTER)
}

func (g *Server) sendAppchain(c *gin.Context, appchainType pb.Message_Type) {
	// target sidecar id
	sidecarID := c.Query("sidecar_id")
	var res pb.Response
	var appchain appchainmgr.Appchain
	if err := c.BindJSON(&appchain); err != nil {
		res.Data = []byte(err.Error())
		c.JSON(http.StatusBadRequest, res)
		return
	}
	data, err := json.Marshal(appchain)
	if err != nil {
		g.logger.Errorln(err)
	}

	msg := pb.Msg(appchainType, true, data)
	ackMsg, err := g.router.Send(sidecarID, msg)
	if err != nil {
		res.Data = []byte(err.Error())
		c.JSON(http.StatusInternalServerError, res)
		return
	}

	g.handleAckAppchain(c, ackMsg)
}

func (g *Server) handleAckAppchain(c *gin.Context, msg *pb.Message) {
	app := appchainmgr.Appchain{}
	if err := json.Unmarshal(msg.Payload.Data, &app); err != nil {
		g.logger.Error(err)
		return
	}

	res := &response{}
	switch msg.Type {
	case pb.Message_APPCHAIN_REGISTER:
		res.Data = []byte(fmt.Sprintf("appchain register successfully, id is %s\n", app.ID))
	case pb.Message_APPCHAIN_UPDATE:
		res.Data = []byte(fmt.Sprintf("appchain update successfully, id is %s\n", app.ID))
	case pb.Message_APPCHAIN_GET:
		res.Data = msg.Payload.Data
	}

	c.JSON(http.StatusOK, res)
}

func (g *Server) getAppchain(c *gin.Context) {
	res := &response{}

	// target sidecar id
	selfSidecarID, err := g.getSelfSidecarID(c)
	if err != nil {
		c.JSON(http.StatusInternalServerError, err.Error())
		return
	}
	targetSidecarID := c.Query("sidecar_id")
	appchain := &appchainmgr.Appchain{
		ID: selfSidecarID,
	}

	data, err := json.Marshal(appchain)
	if err != nil {
		c.JSON(http.StatusInternalServerError, err.Error())
		return
	}
	msg := pb.Msg(pb.Message_APPCHAIN_GET, true, data)

	ackMsg, err := g.router.Send(targetSidecarID, msg)
	if err != nil {
		res.Data = []byte(err.Error())
		c.JSON(http.StatusInternalServerError, res)
		return
	}

	g.handleAckAppchain(c, ackMsg)
}

func (g *Server) getSelfSidecarID(ctx *gin.Context) (string, error) {
	repoRoot := g.config.RepoRoot
	keyPath := filepath.Join(repoRoot, "key.json")

	privKey, err := asym.RestorePrivateKey(keyPath, "link33")
	if err != nil {
		return "", err
	}

	address, err := privKey.PublicKey().Address()
	if err != nil {
		return "", err
	}

	return address.String(), nil
}
